{{
        KEYPAD ENCODER - Peter Jakacki

        Starts up a cog that scans a keypad matrix
        and translates and buffers the codes.
        This code was originally written as a hardcoded 4X4 (tested)
        but was adapted for larger matrices (not tested)
        This is a simple but complete Spin implementation

Keypad actions don't have to be fast but should be reliable so debounce periods are several ms
The keypad matrix is made up of rows (horizontal) and columns. (vertical)
All column lines are driven low while the row lines are inputs pulled high (10K).
When a key is pressed one of the row lines will become low.
The column lines are now scanned one by one to find out which column it belongs to.
When the column is found the column index is combined with the row index to form a keycode.
The keycode is translated in a user specified table and pushed into the keypad buffer.
}}

CON
  bufsz = 16                    ' keypad buffer size (keep in multiples of 2)
  
VAR
  long  stack[32]
  long  rows,cols,colpins,rowpins,akey,lastkey,keyrd,keywr,keycodes,ms1
  byte  keybuf[bufsz]




' Start the keypad scanner - column and row pins do not have to be adjacent 
' If keypad code translation is required then set keycodeptr to address of
' translation table or else set to zero for no translation 
'
pub start(keyrows,keycols,firstcolpin,firstrowpin,keycodeptr) : okay | i
  rows := keyrows
  cols := keycols
  colpins := firstcolpin
  rowpins := firstrowpin
  keyrd := 0
  keywr := 0
  keycodes := keycodeptr                                ' remember pointer to user keycodes
  ms1 := clkfreq / 1000 - 4296
  okay := cognew(scankeys,@stack) +1                    ' run keypad in it's own cog


' Returns a non-zero value if there is a keypad code in the buffer
' 
pub getkey : keycode
  keycode := 0
  if keywr <> keyrd
    keycode := keybuf[keyrd++&(bufsz-1)]
  return

' ******************* background methods run in their own cogs ********************

pri scankeys                                           'main scanning loop
  dira[rowpins..rowpins+rows-1]~                       'ensure row lines are inputs 
  repeat
    outa[colpins..colpins+cols-1]~                     'drive all columns low (all active)
    dira[colpins..colpins+cols-1]~~
    lastkey := 0
    repeat until row                                    'wait until any key is pressed
      ms(5)                                             'oh hum, this is so boring                                
    ms(5)                                               'there was a signal, debounce
    if row                                              'double-check row again (noise?)
      keypressed                                        'seems solid, let's try to encode it      

pri keypressed | i
      outa[colpins..colpins+cols-1]~                    'columns are always active low
      repeat i from 0 to cols-1                         'find out whch column is active
        dira[colpins..colpins+cols-1]~                  'all columns inactive except...
        dira[colpins+i]~~                               'only one column active
        ms(1)                                           'debounce period
        if row                                          'ignore false press (glitch)
          akey := i + (( >|row)-1 )*cols                'convert row and column
          if keycodes                                   'translate keycodes?
            akey := @keycodes.byte[akey]                'lookup new code (usually ASCII)
          if lastkey <> akey                            'great! but is it a new key?
            lastkey := keybuf[keywr++&(bufsz-1)] := akey    'yes, let's buffer it
          release                                       'wait for button release
          return
      return

pri release
  repeat until row == 0                                'row lines are released
    repeat while row                                   'wait if still glitching (shouldn't)
      ms(5)                                            'debounce a further 5ms
  
pri row
  return !ina[rowpins..rowpins+rows-1] & (rows*cols-1)   

pri ms(period)
    waitcnt(ms1*period + cnt)